from outatime.util.agenda import *


def test_get_quarter():
    day = datetime.strptime("2020-01-07", "%Y-%m-%d").date()
    q = get_quarter(day)
    assert q == 1, "Wrong quarter returned, expected 1."
    day = datetime.strptime("2020-05-07", "%Y-%m-%d")
    q = get_quarter(day)
    assert q == 2, "Wrong quarter returned, expected 2."
    day = datetime.strptime("2020-08-07", "%Y-%m-%d")
    q = get_quarter(day)
    assert q == 3, "Wrong quarter returned, expected 3."
    day = datetime.strptime("2020-11-07", "%Y-%m-%d")
    q = get_quarter(day)
    assert q == 4, "Wrong quarter returned, expected 4."


def test_first_day_of_week():
    day = datetime.strptime("2020-01-07", "%Y-%m-%d").date()
    res = first_day_of_week(day)
    assert res == datetime.strptime("2020-01-06", "%Y-%m-%d").date(), "Wrong day returned."


def test_last_day_of_week():
    day = datetime.strptime("2020-01-07", "%Y-%m-%d").date()
    res = last_day_of_week(day)
    assert res == datetime.strptime("2020-01-12", "%Y-%m-%d").date(), "Wrong day returned."


def test_first_day_of_month():
    day = datetime.strptime("2020-01-07", "%Y-%m-%d").date()
    res = first_day_of_month(day)
    assert res == datetime.strptime("2020-01-01", "%Y-%m-%d").date(), "Wrong day returned."


def test_last_day_of_month():
    day = datetime.strptime("2020-01-07", "%Y-%m-%d").date()
    res = last_day_of_month(day)
    assert res == datetime.strptime("2020-01-31", "%Y-%m-%d").date(), "Wrong day returned."


def test_first_day_of_quarter():
    day = datetime.strptime("2020-01-07", "%Y-%m-%d").date()
    res = first_day_of_quarter(day)
    assert res == datetime.strptime("2020-01-01", "%Y-%m-%d").date(), "Wrong day returned."


def test_last_day_of_quarter():
    day = datetime.strptime("2020-01-07", "%Y-%m-%d").date()
    res = last_day_of_quarter(day)
    assert res == datetime.strptime("2020-03-31", "%Y-%m-%d").date(), "Wrong day returned."


def test_first_day_of_year():
    day = datetime.strptime("2020-01-07", "%Y-%m-%d").date()
    res = first_day_of_year(day)
    assert res == datetime.strptime("2020-01-01", "%Y-%m-%d").date(), "Wrong day returned."


def test_last_day_of_year():
    day = datetime.strptime("2020-01-07", "%Y-%m-%d").date()
    res = last_day_of_year(day)
    assert res == datetime.strptime("2020-12-31", "%Y-%m-%d").date(), "Wrong day returned."


def test_days_of_month():
    days = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    for month_n, month_days in enumerate(days):
        day = datetime.strptime(f"2020-{month_n+1}-07", "%Y-%m-%d").date()
        res = days_of_month(day)
        assert res == month_days, f"Wrong number of days in month {month_n+1}."


def test_days_of_quarter():
    day = datetime.strptime("2020-01-07", "%Y-%m-%d").date()
    res = days_of_quarter(day)
    assert res == 91, "Wrong number of days in quarter 1."

    day = datetime.strptime("2020-03-07", "%Y-%m-%d").date()
    res = days_of_quarter(day)
    assert res == 91, "Wrong number of days in quarter 2."

    day = datetime.strptime("2020-08-07", "%Y-%m-%d").date()
    res = days_of_quarter(day)
    assert res == 92, "Wrong number of days in quarter 3."

    day = datetime.strptime("2020-12-07", "%Y-%m-%d").date()
    res = days_of_quarter(day)
    assert res == 92, "Wrong number of days in quarter 3."


def test_days_of_year():
    day = datetime.strptime("2020-01-01", "%Y-%m-%d").date()
    res = days_of_year(day)
    assert res == 366, "Wrong number of days in leap year."

    day = datetime.strptime("2021-01-01", "%Y-%m-%d").date()
    res = days_of_year(day)
    assert res == 365, "Wrong number of days in normal year."


def test_is_business_day():
    day = datetime.strptime("2020-01-06", "%Y-%m-%d").date()
    assert is_business_day(day), "Expected a business day."

    day = datetime.strptime("2020-01-05", "%Y-%m-%d").date()
    assert not is_business_day(day), "Expected a week-end day."

    day = datetime.strptime("2020-12-25", "%Y-%m-%d").date()
    assert not is_business_day(day, holidays=[datetime.strptime("2020-12-25", "%Y-%m-%d").date()]), "Expected a holiday."


def test_weekdays_of_range():
    day_sx = datetime.strptime("2020-01-01", "%Y-%m-%d").date()
    day_dx = datetime.strptime("2020-01-31", "%Y-%m-%d").date()
    n_days, days = weekdays_of_range(day_sx, day_dx, 0)
    assert n_days == 4, "Unexpected number of mondays in given range."

    expected_days = [
        datetime.strptime("2020-01-06", "%Y-%m-%d").date(),
        datetime.strptime("2020-01-13", "%Y-%m-%d").date(),
        datetime.strptime("2020-01-20", "%Y-%m-%d").date(),
        datetime.strptime("2020-01-27", "%Y-%m-%d").date()
    ]

    assert expected_days == days, "Unexpected days returned."


def test_calendar_by_steps():
    day_sx = datetime.strptime("2020-01-01", "%Y-%m-%d").date()
    day_dx = datetime.strptime("2020-01-10", "%Y-%m-%d").date()
    res = calendar_by_steps(day_sx, day_dx, step=relativedelta(days=2))

    expected_days = [
        datetime.strptime("2020-01-01", "%Y-%m-%d").date(),
        datetime.strptime("2020-01-03", "%Y-%m-%d").date(),
        datetime.strptime("2020-01-05", "%Y-%m-%d").date(),
        datetime.strptime("2020-01-07", "%Y-%m-%d").date(),
        datetime.strptime("2020-01-09", "%Y-%m-%d").date()
    ]
    assert expected_days == res, "Unexpected days returned."

    try:
        _ = calendar_by_steps(day_dx, day_sx, step=relativedelta(days=2))
        raise AssertionError("Inverted days exception not caught.")
    except AssertionError:
        pass



import pickle
from dataclasses import dataclass
from datetime import date


@dataclass
class TimeSeriesData:
    day: date
    data: ...

    def __deepcopy__(self, *args):
        return pickle.loads(pickle.dumps(self))



from datetime import datetime

from outatime.granularity.granularity import MonthlyGranularity
from outatime.timeseries.batches import pick_a_weekday
from test.utils import data_generation


def test_pick_a_weekday():
    tsl = data_generation()

    res = pick_a_weekday(
        ts=tsl,
        granularity=MonthlyGranularity(),
        day_of_batch=-1,
        weekday=0
    )
    assert len(res) == 60, "Unexpected number of elements in resulting time series."

    expected_dates = [
        res[0].day == datetime.strptime("2020-01-27", "%Y-%m-%d").date(),
        res[1].day == datetime.strptime("2020-02-24", "%Y-%m-%d").date(),
        res[-2].day == datetime.strptime("2024-11-25", "%Y-%m-%d").date(),
        res[-1].day == datetime.strptime("2024-12-30", "%Y-%m-%d").date()
    ]
    assert all(expected_dates), "Unexpected dates."

    assert all([day.weekday() == 0 for day in res.dates]), "Unexpected days of week in resulting time series."


def test_pick_a_weekday_first():
    tsl = data_generation()

    res = pick_a_weekday(
        ts=tsl,
        granularity=MonthlyGranularity(),
        day_of_batch=0,
        weekday=0
    )
    assert len(res) == 59, "Unexpected number of elements in resulting time series."

    expected_dates = [
        res[0].day == datetime.strptime("2020-02-03", "%Y-%m-%d").date(),
        res[1].day == datetime.strptime("2020-03-02", "%Y-%m-%d").date(),
        res[-2].day == datetime.strptime("2024-11-04", "%Y-%m-%d").date(),
        res[-1].day == datetime.strptime("2024-12-02", "%Y-%m-%d").date()
    ]
    assert all(expected_dates), "Unexpected dates."

    assert all([day.weekday() == 0 for day in res.dates]), "Unexpected days of week in resulting time series."


def test_pick_a_weekday_all_wds():
    tsl = data_generation()

    for i in range(7):
        res = pick_a_weekday(
            ts=tsl,
            granularity=MonthlyGranularity(),
            day_of_batch=0,
            weekday=i
        )
        assert all([day.weekday() == i for day in res.dates]), "Unexpected days of week in resulting time series."



import random
import time

from outatime.dataclass.time_series_data import TimeSeriesData
from outatime.granularity.granularity import *
from outatime.timeseries.time_series import TimeSeries


def data_generation(start_date='2020-01-07', end_date='2025-01-04', step=relativedelta(days=1), empty_data_step=2, empty_data={}):
    z = []
    i = 0
    print("Generating input time series...")
    t = time.time()
    start_date = datetime.strptime(start_date, "%Y-%m-%d").date()
    end_date = datetime.strptime(end_date, "%Y-%m-%d").date()
    while start_date <= end_date:
        i += 1
        data = empty_data
        if i % empty_data_step > 0:
            data = {'pippo': random.randint(200, 300), 'pluto': random.randint(200, 300)}
        day = TimeSeriesData(day=start_date, data=data)
        z.append(day)

        start_date += step

    tsl = TimeSeries(z)
    print(f"*** Executed in: {round(time.time() - t, 4)}s.")
    print("Done.")
    return tsl


def compare(expected_res, res) -> bool:
    return all([x == y for x, y in zip(res, expected_res)])



from datetime import datetime

from outatime.granularity.granularity import MonthlyGranularity
from outatime.timeseries.batches import pick_a_day
from test.utils import data_generation


def test_pick_a_day():
    tsl = data_generation()

    res = pick_a_day(tsl, granularity=MonthlyGranularity())
    assert len(res) == 60, "Unexpected number of elements in resulting time series."

    expected_dates = [
        res[0].day == datetime.strptime("2020-01-31", "%Y-%m-%d").date(),
        res[1].day == datetime.strptime("2020-02-29", "%Y-%m-%d").date(),
        res[-2].day == datetime.strptime("2024-11-30", "%Y-%m-%d").date(),
        res[-1].day == datetime.strptime("2024-12-31", "%Y-%m-%d").date()
    ]
    assert all(expected_dates), "Unexpected dates."


def test_pick_a_day_index():
    tsl = data_generation()

    res = pick_a_day(tsl, granularity=MonthlyGranularity(), day_of_batch=0)
    assert len(res) == 60, "Unexpected number of elements in resulting time series."

    expected_dates = [
        res[0].day == datetime.strptime("2020-02-01", "%Y-%m-%d").date(),
        res[1].day == datetime.strptime("2020-03-01", "%Y-%m-%d").date(),
        res[-2].day == datetime.strptime("2024-12-01", "%Y-%m-%d").date(),
        res[-1].day == datetime.strptime("2025-01-01", "%Y-%m-%d").date()
    ]
    assert all(expected_dates), "Unexpected dates."



import calendar
from datetime import date, datetime, timedelta
from typing import List

from dateutil.rrule import DAILY, rrule, MO, TU, WE, TH, FR, SA, SU
import numpy as np

from .decorators import day_or_datetime
from ..util.relativedelta import relativedelta

WEEKMASK = [
    [1, 0, 0, 0, 0, 0, 0],
    [0, 1, 0, 0, 0, 0, 0],
    [0, 0, 1, 0, 0, 0, 0],
    [0, 0, 0, 1, 0, 0, 0],
    [0, 0, 0, 0, 1, 0, 0],
    [0, 0, 0, 0, 0, 1, 0],
    [0, 0, 0, 0, 0, 0, 1]
]

WEEKDAYS = [MO, TU, WE, TH, FR, SA, SU]


@day_or_datetime
def get_quarter(day: date) -> int:
    """
    Searches the quarter of the year that includes the given day.
    Possible outputs are 1, 2, 3, 4.
    """
    return (day.month - 1) // 3 + 1


@day_or_datetime
def first_day_of_week(day: date) -> date:
    """Searches the first day of the week that includes the given day."""
    return day - relativedelta(days=day.weekday())


@day_or_datetime
def last_day_of_week(day: date) -> date:
    """Searches the last day of the week that includes the given day."""
    return day + relativedelta(days=6 - day.weekday())


@day_or_datetime
def first_day_of_month(day: date) -> date:
    """Searches the first day of the month that includes the given day."""
    return day.replace(day=1)


@day_or_datetime
def last_day_of_month(day: date) -> date:
    """Searches the last day of the month that includes the given day."""
    return day.replace(day=days_of_month(day=day))


@day_or_datetime
def first_day_of_quarter(day: date) -> date:
    """Searches the first day of the quarter that includes the given day."""
    quarter = get_quarter(day)
    return datetime(year=day.year, month=3 * quarter - 2, day=1).date()


@day_or_datetime
def last_day_of_quarter(day: date) -> date:
    """Searches the last day of the quarter that includes the given day."""
    quarter = get_quarter(day)
    return datetime(
        year=day.year + 3 * quarter // 12,
        month=3 * quarter % 12 + 1,
        day=1
    ).date() + timedelta(days=-1)


@day_or_datetime
def first_day_of_year(day: date) -> date:
    """Searches the first day of the year that includes the given day."""
    return day.replace(month=1, day=1)


@day_or_datetime
def last_day_of_year(day: date) -> date:
    """Searches the last day of the year that includes the given day."""
    return day.replace(month=12, day=31)


@day_or_datetime
def days_of_month(day: date) -> int:
    """Returns the length of the month containing the day, expressed in days."""
    return calendar.monthrange(day.year, day.month)[1]


@day_or_datetime
def days_of_quarter(day: date) -> int:
    """
    Returns the length of the quarter containing the day, expressed in days.
    """
    first_day = first_day_of_quarter(day)
    last_day = last_day_of_quarter(day)
    return (last_day - first_day).days + 1


@day_or_datetime
def days_of_year(day: date) -> int:
    """Returns the length of the year containing the day, expressed in days."""
    return 365 + calendar.isleap(day.year)


@day_or_datetime
def is_business_day(day: date, holidays: List[date] = []) -> bool:
    """
    Checks if a given day is a weekday excluding holidays (if given).

    Args:
        day (date): Day to check.
        holidays (List[date], optional): List of holidays dates. 
        Defaults to [].

    Returns:
        bool: True if the day is a business day.
    """
    holidays_strings = [hd.strftime("%Y-%m-%d") for hd in holidays]
    return np.is_busday(day.strftime("%Y-%m-%d"), weekmask='1111100', holidays=holidays_strings)


def weekdays_of_range(start_date: date, end_date: date, weekday: int, exclusive: bool = False):
    """
    Searches for the all the occurrences of a given day of week in
    the given range of dates.

    Args:
        start_date (date):
        end_date (date):
        weekday (int): Number of the day of the week (0-indexed).
        exclusive (bool, optional): Don't include end_date in the
        searched range. Defaults to False.

    Returns:
        int: Number of found days.
        list: List of found days.
    """
    assert 0 <= weekday <= 6, "Weekday must be between 0 and 6."
    day_times = rrule(
        freq=DAILY,
        dtstart=start_date,
        until=end_date + relativedelta(days=int(exclusive)),
        byweekday=(WEEKDAYS[weekday])
    )
    days = [day_time.date() for day_time in day_times]
    n_days = len(days)
    return n_days, days


def calendar_by_steps(start_date: date, end_date: date, step: relativedelta) -> list:
    """
    Evaluates steps between start_date and end_date.

    Args:
        start_date (date): First date of the period.
        end_date (date): Last date of the period.
        step (relativedelta): Time step.

    Returns:
        list: Time range dates with a defined step.
    """

    assert end_date > start_date, "end date must be greater than start date"

    res = []
    while start_date <= end_date:
        res.append(
            start_date
        )
        start_date += step

    return res



from outatime.util.bisect import *


def test_index_of():
    _list = ['a', 'b', 'c', 'd']
    assert index_of(_list, 'b') == 1, "Wrong index returned."
    try:
        _ = index_of(_list, 'g')
        raise AssertionError("Exception not found.")
    except ValueError:
        pass


def test_find_lt():
    _list = ['a', 'b', 'c', 'd']
    assert find_lt(_list, 'b') == 'a', "Wrong element returned."


def test_find_lte():
    _list = ['a', 'b', 'c', 'd']
    assert find_lte(_list, 'b') == 'b', "Wrong element returned."


def test_find_gt():
    _list = ['a', 'b', 'c', 'd']
    assert find_gt(_list, 'b') == 'c', "Wrong element returned."


def test_find_gte():
    _list = ['a', 'b', 'c', 'd']
    assert find_gte(_list, 'b') == 'b', "Wrong element returned."


def test_find_delimiters():
    _list = ['a', 'b', 'c', 'd']
    idx_min, idx_max = find_delimiters(_list, 'b', 'c')
    assert idx_min == 1 and idx_max == 2, "Wrong indexes returned."


from bisect import bisect_left, bisect_right


def index_of(_list: list, value):
    """Find index of element in given list."""
    i = bisect_left(_list, value)
    try:
        assert value == _list[i], "Exact element not found."
        return i
    except (IndexError, AssertionError):
        raise ValueError


def find_lt(_list: list, value):
    """Find rightmost value less than value."""
    i = bisect_left(_list, value)
    if i:
        return _list[i-1]
    raise ValueError


def find_lte(_list: list, value):
    """Find rightmost value less than or equal to value."""
    i = bisect_right(_list, value)
    if i:
        return _list[i-1]
    raise ValueError


def find_gt(_list: list, value):
    """Find leftmost value greater than value."""
    i = bisect_right(_list, value)
    if i != len(_list):
        return _list[i]
    raise ValueError


def find_gte(_list: list, value):
    """Find leftmost item greater than or equal to value."""
    i = bisect_left(_list, value)
    if i != len(_list):
        return _list[i]
    raise ValueError


def find_delimiters(_list: list, first_element, second_element) -> tuple:
    """
    Given a list of elements, searches:
        * the index of the first element included in the given range
        * the index of the last element included in the given range

    Args:
        _list (list): Input data.
        first_element: Low limit of the range to search.
        second_element: High limit of the range to search.

    Returns:
        tuple: First and last indexes of the searched range.
    """
    idx_min = bisect_left(_list, first_element)
    idx_max = bisect_right(_list[idx_min:], second_element) + idx_min - 1
    return idx_min, idx_max



from copy import deepcopy
from datetime import datetime

from outatime.granularity.granularity import MonthlyGranularity, WeeklyGranularity
from outatime.util.relativedelta import relativedelta

from test.utils import data_generation, compare


def test_deepcopy():
    tsl = data_generation()
    x = deepcopy(tsl)
    assert id(x) != id(tsl), "Objects' identities are the same."


def test_cut():
    min_date = datetime.strptime("2020-01-11", "%Y-%m-%d").date()
    max_date = datetime.strptime("2021-01-11", "%Y-%m-%d").date()

    tsl = data_generation()
    y = tsl.cut(min_date=min_date, max_date=max_date, inplace=False)

    assert y[0].day == min_date, "Starting date is wrong."
    assert y[-1].day == max_date, "Ending date is wrong."
    assert tsl[0].day != min_date and tsl[-1].day != max_date, "Original time series changed."


def test_cut_inplace():
    min_date = datetime.strptime("2020-01-11", "%Y-%m-%d").date()
    max_date = datetime.strptime("2021-01-11", "%Y-%m-%d").date()

    tsl = data_generation()
    tsl.cut(min_date=min_date, max_date=max_date, inplace=True)

    assert tsl[0].day == min_date, "Starting date is wrong."
    assert tsl[-1].day == max_date, "Ending date is wrong."


def test_resample():
    tsl = data_generation(start_date='2020-01-01', end_date='2020-04-01', step=relativedelta(months=1))
    res = tsl.resample(granularity=WeeklyGranularity())
    expected_res = [
        datetime.strptime('2019-12-30', '%Y-%m-%d').date(),
        datetime.strptime('2020-01-06', '%Y-%m-%d').date(),
        datetime.strptime('2020-01-13', '%Y-%m-%d').date(),
        datetime.strptime('2020-01-20', '%Y-%m-%d').date(),
        datetime.strptime('2020-01-27', '%Y-%m-%d').date(),
        datetime.strptime('2020-02-03', '%Y-%m-%d').date(),
        datetime.strptime('2020-02-10', '%Y-%m-%d').date(),
        datetime.strptime('2020-02-17', '%Y-%m-%d').date(),
        datetime.strptime('2020-02-24', '%Y-%m-%d').date(),
        datetime.strptime('2020-03-02', '%Y-%m-%d').date(),
        datetime.strptime('2020-03-09', '%Y-%m-%d').date(),
        datetime.strptime('2020-03-16', '%Y-%m-%d').date(),
        datetime.strptime('2020-03-23', '%Y-%m-%d').date(),
        datetime.strptime('2020-03-30', '%Y-%m-%d').date(),
    ]
    assert compare(expected_res, res.dates), "Unexpected resample result."
    assert not compare(expected_res, tsl.dates), "Original data changed."


def test_resample_inplace():
    tsl = data_generation(start_date='2020-01-01', end_date='2020-04-01', step=relativedelta(months=1))
    tsl.resample(granularity=WeeklyGranularity(), inplace=True)
    expected_res = [
        datetime.strptime('2019-12-30', '%Y-%m-%d').date(),
        datetime.strptime('2020-01-06', '%Y-%m-%d').date(),
        datetime.strptime('2020-01-13', '%Y-%m-%d').date(),
        datetime.strptime('2020-01-20', '%Y-%m-%d').date(),
        datetime.strptime('2020-01-27', '%Y-%m-%d').date(),
        datetime.strptime('2020-02-03', '%Y-%m-%d').date(),
        datetime.strptime('2020-02-10', '%Y-%m-%d').date(),
        datetime.strptime('2020-02-17', '%Y-%m-%d').date(),
        datetime.strptime('2020-02-24', '%Y-%m-%d').date(),
        datetime.strptime('2020-03-02', '%Y-%m-%d').date(),
        datetime.strptime('2020-03-09', '%Y-%m-%d').date(),
        datetime.strptime('2020-03-16', '%Y-%m-%d').date(),
        datetime.strptime('2020-03-23', '%Y-%m-%d').date(),
        datetime.strptime('2020-03-30', '%Y-%m-%d').date(),
    ]
    assert compare(expected_res, tsl.dates), "Unexpected resample result."



from ..dataclass.time_series_data import TimeSeriesData
from ..timeseries.time_series import TimeSeries


def take_first_available(a, b):
    if a is not None:
        return a
    elif b is not None:
        return b
    else:
        raise Exception("Both arguments are None.")


def intersection(tsl_a: TimeSeries, tsl_b: TimeSeries, conflict_method=take_first_available) -> TimeSeries:
    """
    Given two input time series, generates a new time series with only shared 
    days and all the contained values.

    Args:
        tsl_a (TimeSeries): First input time series.
        tsl_b (TimeSeries): Second input time series.
        conflict_method (None, optional): Method to apply when choosing
        data for matching days. Defaults to take_first_available.

    Returns:
        TimeSeries: Output timeseries with shared days.
    """
    intersection_dates = list(set(tsl_a.dates) & set(tsl_b.dates))
    intersection_dates.sort()
    intersection_result = []
    for int_date in intersection_dates:
        intersection_result.append(
            TimeSeriesData(
                day=int_date,
                data=conflict_method(tsl_a.get(day=int_date).data, tsl_b.get(day=int_date).data)
            )
        )
    return TimeSeries(intersection_result)


def union(tsl_a: TimeSeries, tsl_b: TimeSeries, conflict_method=take_first_available) -> TimeSeries:
    """
    Given two input time series, generates a new time series with the union of 
    all days of both series.

    Args:
        tsl_a (TimeSeries): First input time series.
        tsl_b (TimeSeries): Second input time series.
        conflict_method (None, optional): Method to apply when choosing
        data for matching days. Defaults to take_first_available.

    Returns:
        TimeSeries: Output timeseries with all days.
    """
    union_dates = list(set(tsl_a.dates) | set(tsl_b.dates))
    union_dates.sort()
    union_result = []
    for uni_date in union_dates:
        union_result.append(
            TimeSeriesData(
                day=uni_date,
                data=conflict_method(tsl_a.get(day=uni_date).data, tsl_b.get(day=uni_date).data)
            )
        )
    return TimeSeries(union_result)



from datetime import datetime


def day_or_datetime(func):
    """
    Decorator that transform a datetime object to a date object when it is 
    given as input to the decorated function.
    """
    def _check_if_datetime(day, *args, **kwargs):
        if isinstance(day, datetime):
            return func(day=day.date(), *args, **kwargs)
        return func(day=day,  *args, **kwargs)
    return _check_if_datetime



from datetime import datetime

from outatime.granularity.granularity import MonthlyGranularity
from outatime.timeseries.batches import split
from test.utils import data_generation


def test_split():
    tsl = data_generation()
    res = split(tsl, granularity=MonthlyGranularity())
    assert len(res) == 61, "Unexpected number of batches."
    expected_dates = [
        res[0][0].day == datetime.strptime("2020-01-07", "%Y-%m-%d").date(),
        res[0][-1].day == datetime.strptime("2020-01-31", "%Y-%m-%d").date(),
        res[-1][0].day == datetime.strptime("2025-01-01", "%Y-%m-%d").date(),
        res[-1][-1].day == datetime.strptime("2025-01-04", "%Y-%m-%d").date()
    ]
    assert all(expected_dates), "Unexpected dates."


def test_split_indexed():
    tsl = data_generation()
    res = split(tsl, granularity=MonthlyGranularity(), first_day_of_batch=5, last_day_of_batch=10,)
    assert len(res) == 60, "Unexpected number of batches."
    expected_dates = [
        res[0][0].day == datetime.strptime("2020-01-07", "%Y-%m-%d").date(),
        res[0][-1].day == datetime.strptime("2020-01-11", "%Y-%m-%d").date(),
        res[-1][0].day == datetime.strptime("2024-12-06", "%Y-%m-%d").date(),
        res[-1][-1].day == datetime.strptime("2024-12-11", "%Y-%m-%d").date()
    ]
    assert all(expected_dates), "Unexpected dates."


def test_split_drop_tails():
    tsl = data_generation()
    res = split(tsl, granularity=MonthlyGranularity(), drop_tails=True)
    assert len(res) == 59, "Unexpected number of batches."
    expected_dates = [
        res[0][0].day == datetime.strptime("2020-02-01", "%Y-%m-%d").date(),
        res[0][-1].day == datetime.strptime("2020-02-29", "%Y-%m-%d").date(),
        res[-1][0].day == datetime.strptime("2024-12-01", "%Y-%m-%d").date(),
        res[-1][-1].day == datetime.strptime("2024-12-31", "%Y-%m-%d").date()
    ]
    assert all(expected_dates), "Unexpected dates."



from .granularity import *


class GranularityFactory:
    @staticmethod
    def get_granularity(granularity):
        if granularity in ('daily', 'D'):
            return DailyGranularity
        elif granularity in ('weekly', 'W'):
            return WeeklyGranularity
        elif granularity in ('monthly', 'M'):
            return MonthlyGranularity
        elif granularity in ('quarterly', 'Q'):
            return QuarterlyGranularity
        elif granularity in ('yearly', 'Y'):
            return YearlyGranularity
        else:
            raise ValueError(f'Invalid granularity value: {granularity}')



from datetime import date
from abc import abstractmethod, ABC

from ..util.agenda import *
from ..util.relativedelta import relativedelta

assertion_err_msg = "Given number exceeds the time range limit."


class Granularity(ABC):
    """
    Abstract class used to manage time ranges of different lengths.
    """
    delta: relativedelta

    @abstractmethod
    def get_beginning_of_granularity(self, day: date) -> date:
        """
        Move the given date to the beginning of its granularity period.

        Example:
            (with a Monthly granularity)
            day = 2022-04-16
            returns 2022-04-01

        Args:
            day (date): Input date.

        Returns:
            date: First day of the granularity step.
        """
        pass

    @abstractmethod
    def get_end_of_granularity(self, day: date) -> date:
        """
        Move the given date to the end of its granularity period.

        Example:
            (with a Monthly granularity)
            day = 2022-04-16
            returns 2022-04-30

        Args:
            day (date): Input date.

        Returns:
            date: Last day of the granularity step.
        """
        pass

    @abstractmethod
    def assert_included_day(self, day: date, days: int):
        """
        Checks if a range of days is included in the number of days of 
        the granularity step for the given day.

        Args:
            day (date): Reference day for the granularity step.
            days (int): Number of the days to check.
        """
        pass

    def get_n_day_of_granularity(self, day: date, idx: int) -> date:
        """
        Move the given date to N-th day of its granularity period.

        Example:
            (with a Monthly granularity)
            day = 2022-04-16
            n = 8
            returns 2022-04-08

        Args:
            day (date): Input date.
            idx (int): Number of the day to retrieve (0-indexed).

        Returns:
            date: N-th day of the granularity step.
        """
        self.assert_included_day(day=day, days=idx)
        if idx == -1:
            return self.get_end_of_granularity(day)
        return self.get_beginning_of_granularity(day) + relativedelta(days=idx)

    def get_n_weekday_of_granularity(self, day: date, weekday: int, idx: int) -> date:
        """
        Move the given date to N-th required day of the week of its granularity 
        period.

        Example:
            (with Monthly Granularity)
            day: 2020-04-19
            weekday: 3 (Wednesday)
            idx: 2 (the 3rd one starting from zero)

            returns 2020-04-20 (the 3rd Wednesday of the month)

        Args:
            day (date): Input date.
            weekday (int): Number of the day of the week (0-indexed).
            idx (int): Number of the day to retrieve (0-indexed).

        Returns:
            date: N-th day of the week of the granularity step.
        """
        assert 0 <= weekday <= 6, "Weekday must be between 0 and 6."
        first_day = self.get_beginning_of_granularity(day)
        last_day = self.get_end_of_granularity(day)
        n_days, days = weekdays_of_range(first_day, last_day, weekday)
        assert -1 <= idx < n_days, "Index must be between -1 and the total number of the specific weekday in the granularity."
        return days[idx]


class YearlyGranularity(Granularity):

    delta = relativedelta(months=12)

    def get_beginning_of_granularity(self, day: date) -> date:
        return first_day_of_year(day)

    def get_end_of_granularity(self, day: date) -> date:
        return last_day_of_year(day)

    def assert_included_day(self, day: date, days: int):
        assert -1 <= days < days_of_year(day), assertion_err_msg


class QuarterlyGranularity(Granularity):

    delta = relativedelta(months=3)

    def get_beginning_of_granularity(self, day: date) -> date:
        return first_day_of_quarter(day)

    def get_end_of_granularity(self, day: date) -> date:
        return last_day_of_quarter(day)

    def assert_included_day(self, day: date, days: int):
        assert -1 <= days < days_of_quarter(day), assertion_err_msg


class MonthlyGranularity(Granularity):

    delta = relativedelta(months=1)

    def get_beginning_of_granularity(self, day: date) -> date:
        return first_day_of_month(day)

    def get_end_of_granularity(self, day: date) -> date:
        return last_day_of_month(day)

    def assert_included_day(self, day: date, days: int):
        assert -1 <= days < days_of_month(day), assertion_err_msg


class WeeklyGranularity(Granularity):

    delta = relativedelta(weeks=1)

    def get_beginning_of_granularity(self, day: date) -> date:
        return first_day_of_week(day)

    def get_end_of_granularity(self, day: date) -> date:
        return last_day_of_week(day)

    def assert_included_day(self, day: date, days: int):
        assert -1 <= days < 7, assertion_err_msg


class DailyGranularity(Granularity):

    delta = relativedelta(days=1)

    def assert_included_day(self, *args, **kwargs):
        raise AttributeError("Not available method for Daily Granularity.")

    def get_n_weekday_of_granularity(self, *args, **kwargs):
        raise AttributeError("Not available method for Daily Granularity.")

    def get_n_day_of_granularity(self, day: date, *args, **kwargs) -> date:
        return day

    def get_beginning_of_granularity(self, day: date) -> date:
        return day

    def get_end_of_granularity(self, day: date) -> date:
        return day



from datetime import datetime

from outatime.granularity.granularity import MonthlyGranularity, QuarterlyGranularity, YearlyGranularity
from outatime.timeseries.batches import aggregate
from outatime.util.relativedelta import relativedelta
from test.utils import data_generation


def take_first(list_):
    if list_:
        return list_[0]
    return None


def take_last(list_):
    if list_:
        return list_[-1]
    return None


def test_aggregate():
    tsl = data_generation()

    res = aggregate(tsl, granularity=MonthlyGranularity(), method=take_first)
    assert len(res) == 61, "Unexpected number of elements in resulting time series."

    expected_dates = [
        res[0].day == datetime.strptime("2020-01-01", "%Y-%m-%d").date(),
        res[1].day == datetime.strptime("2020-02-01", "%Y-%m-%d").date(),
        res[-2].day == datetime.strptime("2024-12-01", "%Y-%m-%d").date(),
        res[-1].day == datetime.strptime("2025-01-01", "%Y-%m-%d").date()
    ]
    assert all(expected_dates), "Unexpected dates."

    expected_data = [
        res[0].data == tsl[0].data,
        res[1].data == tsl.get(datetime.strptime("2020-02-01", "%Y-%m-%d").date()).data,
        res[-2].data == tsl.get(datetime.strptime("2024-12-01", "%Y-%m-%d").date()).data,
        res[-1].data == tsl.get(datetime.strptime("2025-01-01", "%Y-%m-%d").date()).data,
    ]
    assert all(expected_data), "Unexpected results."


def test_aggregate_on_last():
    tsl = data_generation()

    res = aggregate(tsl, granularity=MonthlyGranularity(), method=take_last)
    assert len(res) == 61, "Unexpected number of elements in resulting time series."

    expected_dates = [
        res[0].day == datetime.strptime("2020-01-01", "%Y-%m-%d").date(),
        res[1].day == datetime.strptime("2020-02-01", "%Y-%m-%d").date(),
        res[-2].day == datetime.strptime("2024-12-01", "%Y-%m-%d").date(),
        res[-1].day == datetime.strptime("2025-01-01", "%Y-%m-%d").date()
    ]
    assert all(expected_dates), "Unexpected dates."

    expected_data = [
        res[0].data == tsl.get(datetime.strptime("2020-01-31", "%Y-%m-%d").date()).data,
        res[1].data == tsl.get(datetime.strptime("2020-02-29", "%Y-%m-%d").date()).data,
        res[-2].data == tsl.get(datetime.strptime("2024-12-31", "%Y-%m-%d").date()).data,
        res[-1].data == tsl[-1].data
    ]
    assert all(expected_data), "Unexpected results."


def test_aggregate_indexed():
    tsl = data_generation()

    res = aggregate(tsl, granularity=MonthlyGranularity(), first_day_of_batch=5, last_day_of_batch=10, method=take_first)
    assert len(res) == 60, "Unexpected number of elements in resulting time series."

    expected_dates = [
        res[0].day == datetime.strptime("2020-01-01", "%Y-%m-%d").date(),
        res[1].day == datetime.strptime("2020-02-01", "%Y-%m-%d").date(),
        res[-2].day == datetime.strptime("2024-11-01", "%Y-%m-%d").date(),
        res[-1].day == datetime.strptime("2024-12-01", "%Y-%m-%d").date()
    ]
    assert all(expected_dates), "Unexpected dates."

    expected_data = [
        res[0].data == tsl[0].data,
        res[1].data == tsl.get(datetime.strptime("2020-02-06", "%Y-%m-%d").date()).data,
        res[-2].data == tsl.get(datetime.strptime("2024-11-06", "%Y-%m-%d").date()).data,
        res[-1].data == tsl.get(datetime.strptime("2024-12-06", "%Y-%m-%d").date()).data,
    ]
    assert all(expected_data), "Unexpected results."


def test_aggregate_store_day():
    tsl = data_generation()

    res = aggregate(tsl, granularity=MonthlyGranularity(), method=take_first, store_day_of_batch=8)
    assert len(res) == 61, "Unexpected number of elements in resulting time series."

    expected_dates = [
        res[0].day == datetime.strptime("2020-01-09", "%Y-%m-%d").date(),
        res[1].day == datetime.strptime("2020-02-09", "%Y-%m-%d").date(),
        res[-2].day == datetime.strptime("2024-12-09", "%Y-%m-%d").date(),
        res[-1].day == datetime.strptime("2025-01-09", "%Y-%m-%d").date()
    ]
    assert all(expected_dates), "Unexpected dates."

    expected_data = [
        res[0].data == tsl[0].data,
        res[1].data == tsl.get(datetime.strptime("2020-02-01", "%Y-%m-%d").date()).data,
        res[-2].data == tsl.get(datetime.strptime("2024-12-01", "%Y-%m-%d").date()).data,
        res[-1].data == tsl.get(datetime.strptime("2025-01-01", "%Y-%m-%d").date()).data,
    ]
    assert all(expected_data), "Unexpected results."


def test_aggregate_drop_tails():
    tsl = data_generation()

    res = aggregate(tsl, granularity=MonthlyGranularity(), first_day_of_batch=5, last_day_of_batch=10, method=take_first, drop_tails=True)
    assert len(res) == 59, "Unexpected number of batches."

    expected_dates = [
        res[0].day == datetime.strptime("2020-02-01", "%Y-%m-%d").date(),
        res[1].day == datetime.strptime("2020-03-01", "%Y-%m-%d").date(),
        res[-2].day == datetime.strptime("2024-11-01", "%Y-%m-%d").date(),
        res[-1].day == datetime.strptime("2024-12-01", "%Y-%m-%d").date()
    ]
    assert all(expected_dates), "Unexpected dates."

    expected_data = [
        res[0].data == tsl.get(datetime.strptime("2020-02-06", "%Y-%m-%d").date()).data,
        res[1].data == tsl.get(datetime.strptime("2020-03-06", "%Y-%m-%d").date()).data,
        res[-2].data == tsl.get(datetime.strptime("2024-11-06", "%Y-%m-%d").date()).data,
        res[-1].data == tsl.get(datetime.strptime("2024-12-06", "%Y-%m-%d").date()).data,
    ]
    assert all(expected_data), "Unexpected results."



from dateutil import relativedelta as rd


class relativedelta(rd.relativedelta):
    """
    Inherits dateutil.relativedelta to add needed properties.
    """
    @property
    def __total_rounded_days__(self):
        return (self.years * 365) + (self.months * 30) + self.days

    @property
    def total_months(self):
        return (self.years * 12) + self.months

    @property
    def total_years(self):
        return self.years + self.months/12 + self.days/365

    def __lt__(self, other):
        if type(other) != type(self): raise TypeError(f"'<' not supported between instances of '{self.__class__.__name__}' and '{other.__class__.__name__}'")
        return self.__total_rounded_days__ < other.__total_rounded_days__

    def __gt__(self, other):
        if type(other) != type(self): raise TypeError(f"'>' not supported between instances of '{self.__class__.__name__}' and '{other.__class__.__name__}'")
        return self.__total_rounded_days__ > other.__total_rounded_days__

    def __le__(self, other):
        if type(other) != type(self): raise TypeError(f"'>=' not supported between instances of '{self.__class__.__name__}' and '{other.__class__.__name__}'")
        return self.__total_rounded_days__ <= other.__total_rounded_days__

    def __ge__(self, other):
        if type(other) != type(self): raise TypeError(f"'<=' not supported between instances of '{self.__class__.__name__}' and '{other.__class__.__name__}'")
        return self.__total_rounded_days__ >= other.__total_rounded_days__

    def __eq__(self, other):
        if type(other) != type(self): raise TypeError(f"'==' not supported between instances of '{self.__class__.__name__}' and '{other.__class__.__name__}'")
        return self.__total_rounded_days__ == other.__total_rounded_days__



from copy import deepcopy
from typing import List
from ..dataclass.time_series_data import TimeSeriesData
from ..granularity.granularity import Granularity, WeeklyGranularity
from ..granularity.utils import get_first_available_beginning
from ..timeseries.time_series import TimeSeries
from ..util.bisect import *


def aggregate(
        ts: TimeSeries,
        method,
        granularity: Granularity = WeeklyGranularity(),
        first_day_of_batch: int = 0,
        last_day_of_batch: int = -1,
        drop_tails: bool = False,
        store_day_of_batch: int = 0,
) -> TimeSeries:
    """
    Divides the input time series in many sub-sets for each contained time step 
    of the given granularity. Then aggregates each sub-set data into a single 
    TimeSeriesData to generate a new TimeSeries output.

    Args:
        ts (TimeSeries): Input time series.
        granularity (Granularity, optional): Time step to use to divide the 
        input series. Defaults to WeeklyGranularity().
        method: Aggregation function to apply.
        first_day_of_batch (int, optional): The day of the time step to use as 
        first delimiter (0-indexed). Defaults to 0.
        last_day_of_batch (int, optional): The day of the time step to use as
        last delimiter (0-indexed). Defaults to -1.
        drop_tails (bool, optional): Choose to remove initial and final days if
        the granularity step is not complete.
        store_day_of_batch (int, optional): The day of the time step to store
        aggregated data into (0-indexed). Defaults to 0.

    Returns:
        TimeSeries: A new time series with aggregated values.
    """
    assert first_day_of_batch >= 0, "'first_day_of_batch' can't be lesser than 0."
    assert last_day_of_batch >= -1 and last_day_of_batch != 0, "'last_day_of_batch' can't be lesser than -1 or equal to 0."

    res = []

    def __next_batch_beginning(day):
        if drop_tails:
            fst_av_beg = get_first_available_beginning(
                day=day,
                input_granularity=ts.data_granularity,
                output_granularity=granularity
            )
        else:
            fst_av_beg = day
        return granularity.get_n_day_of_granularity(fst_av_beg, first_day_of_batch)

    end_date = ts.end_date if drop_tails else granularity.get_end_of_granularity(ts.end_date)

    batch_beginning = __next_batch_beginning(ts.start_date)
    batch_end = granularity.get_n_day_of_granularity(day=batch_beginning, idx=last_day_of_batch)

    temp_ts = deepcopy(ts)

    while batch_end <= end_date:
        reference_day = granularity.get_n_day_of_granularity(day=batch_beginning, idx=store_day_of_batch)

        idx_min, idx_max = find_delimiters(temp_ts.dates, batch_beginning, batch_end)

        subset = temp_ts[idx_min:idx_max+1]
        if len(subset) > 0:
            res.append(
                TimeSeriesData(
                    day=reference_day,
                    data=method([element.data for element in temp_ts[idx_min:idx_max+1]])
                )
            )

        temp_ts = temp_ts[idx_max+1:]
        batch_beginning += granularity.delta
        batch_beginning = granularity.get_n_day_of_granularity(batch_beginning, first_day_of_batch)
        batch_end = granularity.get_n_day_of_granularity(day=batch_beginning, idx=last_day_of_batch)

    return TimeSeries(res)


def pick_a_day(
        ts: TimeSeries,
        granularity: Granularity = WeeklyGranularity(),
        day_of_batch: int = -1
) -> TimeSeries:
    """
    Divides the input time series in many sub-sets for each contained time step 
    of the given granularity. Then returns a new TimeSeries with only the n-th
    day of each batch.

    Args:
        ts (TimeSeries): Input time series.
        granularity (Granularity, optional): Time step to use to divide the 
        input series. Defaults to WeeklyGranularity().
        day_of_batch (int, optional): The day of the time step to retrieve
        (0-indexed). Defaults to -1.

    Returns:
        TimeSeries: A new time series with only a day for each step.
    """
    assert day_of_batch >= -1, "'day_of_batch' can't be lesser than -1."

    res = []

    f_day = granularity.get_n_day_of_granularity(ts.start_date, day_of_batch)
    if f_day < ts.start_date:
        f_day += granularity.delta

    while f_day <= ts.end_date:
        res.append(
            deepcopy(ts.get(day=f_day))
        )

        f_day += granularity.delta
        f_day = granularity.get_n_day_of_granularity(f_day, day_of_batch)

    return TimeSeries(res)


def pick_a_weekday(
        ts: TimeSeries,
        granularity: Granularity = WeeklyGranularity(),
        day_of_batch: int = -1,
        weekday: int = 0
) -> TimeSeries:
    """
    Divides the input time series in many sub-sets for each contained time step
    of the given granularity. Then returns a new TimeSeries with only the n-th 
    chosen day of the week of each batch.

    Args:
        ts (TimeSeries): Input time series.
        granularity (Granularity, optional): Time step to use to divide the
        input series. Defaults to WeeklyGranularity().
        day_of_batch (int, optional): The weekday of the time step to retrieve
        (0-indexed). Defaults to -1.
        weekday (int, optional): The day of the time step to use as
        first delimiter (1-indexed). Defaults to 1.

    Returns:
        TimeSeries: A new time series with only a day for each step.
    """
    assert day_of_batch >= -1, "'day_of_batch' can't be lesser than -1."

    res = []
    f_day = granularity.get_n_weekday_of_granularity(day=ts.start_date, weekday=weekday, idx=day_of_batch)
    if f_day < ts.start_date:
        f_day = granularity.get_n_weekday_of_granularity(day=ts.start_date+granularity.delta, weekday=weekday, idx=day_of_batch)

    while f_day <= ts.end_date:
        res.append(
            deepcopy(ts.get(day=f_day, value=None))
        )

        f_day += granularity.delta
        f_day = granularity.get_n_weekday_of_granularity(day=f_day, weekday=weekday, idx=day_of_batch)

    return TimeSeries(res)


def split(
        ts: TimeSeries,
        granularity: Granularity = WeeklyGranularity(),
        first_day_of_batch: int = 0,
        last_day_of_batch: int = -1,
        drop_tails: bool = False
) -> List[TimeSeries]:
    """
    Divides the input time series in many sub-sets for each contained time step
    of the given granularity.

    Args:
        ts (TimeSeries): Input time series.
        granularity (Granularity, optional): Time step to use to divide the
        input series. Defaults to WeeklyGranularity().
        first_day_of_batch (int, optional): The day of the time step to use as
        first delimiter (0-indexed). Defaults to 0.
        last_day_of_batch (int, optional): The day of the time step to use as
        last delimiter (0-indexed). Defaults to -1.
        drop_tails (bool, optional): Choose to remove initial and final days if
        the granularity step is not complete.

    Returns:
        List[TimeSeries]: A list of smaller time series.
    """
    assert first_day_of_batch >= 0, "'first_day_of_batch' can't be lesser than 0."
    assert last_day_of_batch >= -1 and last_day_of_batch != 0, "'last_day_of_batch' can't be lesser than -1 or equal to 0."

    res = []

    def __next_batch_beginning(day):
        if drop_tails:
            fst_av_beg = get_first_available_beginning(
                day=day,
                input_granularity=ts.data_granularity,
                output_granularity=granularity
            )
        else:
            fst_av_beg = day
        return granularity.get_n_day_of_granularity(fst_av_beg, first_day_of_batch)

    end_date = ts.end_date if drop_tails else granularity.get_end_of_granularity(ts.end_date)

    batch_beginning = __next_batch_beginning(ts.start_date)
    batch_end = granularity.get_n_day_of_granularity(day=batch_beginning, idx=last_day_of_batch)

    temp_ts = deepcopy(ts)

    while batch_end <= end_date:
        idx_min, idx_max = find_delimiters(temp_ts.dates, batch_beginning, batch_end)

        subset = temp_ts[idx_min:idx_max+1]
        if len(subset) > 0:
            res.append(
                subset
            )

        temp_ts = temp_ts[idx_max+1:]
        batch_beginning += granularity.delta
        batch_beginning = granularity.get_n_day_of_granularity(batch_beginning, first_day_of_batch)
        batch_end = granularity.get_n_day_of_granularity(day=batch_beginning, idx=last_day_of_batch)

    return res



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
    long_description = long_description.split("</p>")[1]
    long_description = "# OUTATIME" + long_description

setuptools.setup(
    name='outatime',
    packages=setuptools.find_packages(),
    version='1.0.7',
    description='Python framework to manage time series.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='SynStratos',
    author_email='synstratos.dev@gmail.com',
    url='https://github.com/SynStratos/outatime',
    python_requires='>=3.8, <3.11',
    include_package_data=True,
    classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
    ],
    keywords=['time series', 'ts', 'timeseries', 'temporal data'],
    install_requires=[
        'python-dateutil',
        'statsmodels',
        'numpy',
        'scipy'
    ]
)



import pickle
from datetime import date
from functools import cached_property
from typing import List

from ..dataclass.time_series_data import TimeSeriesData
from ..granularity.granularity import *
from ..util.bisect import index_of, find_delimiters


def get_day(x):
    return x.day


class TimeSeries(List[TimeSeriesData]):
    """
    Class that inherits list to add useful methods for time series management.
    It contains only TimeSeriesData objects as elements.
    """
    data_granularity: Granularity

    def __init__(self, data=[], data_granularity: Granularity = DailyGranularity()):
        super().__init__(data)
        self.data_granularity = data_granularity
        self.__sort__()

    def __add__(self, other):
        super().__add__(other)
        self.__refresh()

    def __deepcopy__(self, *args):
        return pickle.loads(pickle.dumps(self))

    def __delitem__(self, idx):
        super().__delitem__(idx)
        self.__refresh()

    def __getitem__(self, item):
        if isinstance(item, slice):
            return self.__class__(super(TimeSeries, self).__getitem__(item))
        else:
            return super(TimeSeries, self).__getitem__(item)

    def __setitem__(self, key, value):
        super(TimeSeries, self).__setitem__(key, value)
        self.__refresh()

    def __sort__(self):
        """Sort the timeseries by date."""
        self.sort(key=get_day)

    def __clear_cache(self):
        """Clear all cached properties."""
        if hasattr(self, 'dates'):
            del self.dates

    def __refresh(self):
        """Sort the timeseries by date and reset its properties."""
        self.__sort__()
        self.__clear_cache()

    @property
    def start_date(self) -> date:
        """First date of the time series."""
        return self[0].day

    @property
    def end_date(self) -> date:
        """Last date of the time series."""
        return self[-1].day

    @cached_property
    def dates(self):
        """List of all dates available in the time series."""
        return [data.day for data in self]

    def append(self, object_: TimeSeriesData):
        """Add a new TimeSeriesData object to the time series."""
        assert isinstance(object_, TimeSeriesData), "Only TimeSeriesData objects can be appended."
        try:
            self.delete(day=object_.day)
        except ValueError:
            pass
        finally:
            super(TimeSeries, self).append(object_)
            self.__refresh()

    def copy(self):
        """Return a deep copy of the list."""
        return self.__deepcopy__()

    def cut(self, min_date: date, max_date: date, inplace: bool = False):
        """
        Cut the time series selecting only the available days between the
        input delimiters.

        Args:
            min_date (date): Minimum date of range.
            max_date (date): Maximum date of range.
            inplace (bool, optional): Original time series is overwritten
            if set to True. Defaults to False.
        """
        idx_min, idx_max = find_delimiters(self.dates, min_date, max_date)

        if inplace:
            self[:] = self[idx_min:idx_max+1]
            self.__clear_cache()
        else:
            return self.__deepcopy__()[idx_min:idx_max+1]

    def delete(self, day: date):
        """Delete item by day."""
        idx = index_of(self.dates, day)
        self.__delitem__(idx)

    def get(self, day: date, value: ... = None) -> TimeSeriesData:
        """
        Search the time series element for the given day.

        Args:
            day (date): The day to search in the time series.
            value (None, optional): Give a default value to set as 'data' when
            the day is not found. Defaults to None.

        Returns:
            TimeSeriesData: A time series element for the searched day.
        """
        try:
            return self[index_of(self.dates, day)]
        except (KeyError, ValueError):
            return TimeSeriesData(day=day, data=value)

    def resample(self,
                 granularity: Granularity = DailyGranularity(),
                 index_of_granularity: int = 0,
                 inplace: bool = False,
                 default_data: ... = None
                 ):
        """
        Select only needed days for the given granularity.
        If a day is missing, create a new TimeSeriesData with empty series.

        Example:
            [TimeSeriesData(day=2022-04-14, data={'a': 1, 'b': 8}),
            TimeSeriesData(day=2022-04-16, data={'a': 3, 'b': 8})]

            granularity = DailyGranularity()

            Returns:
                [TimeSeriesData(day=2022-04-14, data={'a': 1, 'b': 8}),
                TimeSeriesData(day=2022-04-15, data=None),
                TimeSeriesData(day=2022-04-16, data={'a': 3, 'b': 8})]

        Args:
            granularity (Granularity, optional): Time step to use for
            selecting ranges. Defaults to DailyGranularity().
            index_of_granularity (int, optional): The day of the time step
            to pick as reference (0-indexed). Defaults to 0.
            inplace (bool, optional): Original time series is overwritten
            if set to True. Defaults to False.
            default_data (None, optional): Data to store in days that
            can't be found. Defaults to None.
        """
        f_day = granularity.get_n_day_of_granularity(
            day=self.start_date,
            idx=index_of_granularity
        )
        resampled = []
        temp_ts = self.__deepcopy__()

        resampling_end_date = granularity.get_end_of_granularity(self.end_date)

        while f_day <= resampling_end_date:
            resampled.append(
                temp_ts.get(f_day, value=default_data)
            )
            f_day += granularity.delta
            f_day = granularity.get_n_day_of_granularity(
                day=f_day,
                idx=index_of_granularity
            )

        if inplace:
            self[:] = resampled
            self.data_granularity = granularity
        else:
            return self.__class__(resampled, data_granularity=granularity)

    def update(self, __list: List[TimeSeriesData]):
        """
        Add all elements of the given list of TimeSeriesData to the time series.
        Updates existing elements if already in the time series.

        Args:
            __list (List[TimeSeriesData]): Input list of new elements.
        """
        for new_element in __list:
            self.append(new_element)



from datetime import date

from .granularity import Granularity


def get_first_available_beginning(
        input_granularity: Granularity,
        output_granularity: Granularity,
        day: date
) -> date:
    """
    Searches for the first (output) granularity step beginning
    that includes a given day.

    Args:
        input_granularity (Granularity): Granularity of input data.
        output_granularity (Granularity): Required output granularity.
        day (date): Starting day.

    Returns:
        date: Beginning of the needed granularity.
    """
    beg = output_granularity.get_beginning_of_granularity(day)
    if input_granularity.delta < output_granularity.delta:
        if beg >= day:
            return beg
        else:
            return beg + output_granularity.delta
    elif input_granularity.delta == output_granularity.delta:
        return beg
    else:
        raise TypeError("""Can't compare an input_granularity to an output_granularity with a greater delta.
        It is suggested to resample input data to fit the required granularity.""")



from outatime.timeseries.expr import union, intersection

from test.utils import data_generation, compare


def test_union():
    ts_sx = data_generation()[:10]
    ts_dx = data_generation()[5:15]

    res = union(ts_sx, ts_dx)
    expected_res = list(ts_sx) + list(ts_dx[10:])
    assert len(res) == 15, "Unexpected length of union result."
    assert compare(expected_res, res), "Unexpected union result."


def test_intersection():
    ts_sx = data_generation(empty_data=None)[:10]
    ts_dx = data_generation(empty_data_step=3)[5:15]

    res = intersection(ts_sx, ts_dx)
    expected_res = [ts_dx[0], ts_sx[6], ts_dx[2], ts_sx[8], ts_dx[4]]
    assert len(res) == 5, "Unexpected length of intersection result."
    assert compare(expected_res, res), "Unexpected intersection result."
